import os
import sys
from glob import glob

import numpy
from setuptools import Extension

LINUX_OPTS = {
    "extra_link_args": [
        "-fopenmp",
    ],
    "extra_compile_args": [
        "-fopenmp",
        "-Wno-strict-prototypes",
        "-Wno-maybe-uninitialized",
        "-O3",
        "-std=c99",
    ],
}
OSX_OPTS = {
    "extra_link_args": [
        "-L/usr/local/opt/libomp/lib",
        "-fopenmp",
    ],
    "extra_compile_args": [
        "-Wno-strict-prototypes",
        "-Wno-uninitialized",
        "-fopenmp",
        "-O3",
    ],
}
WINDOWS_OPTS = {
    "extra_compile_args": [
        "/Ox",
    ]
}


if os.name == "posix":
    if sys.platform.startswith("darwin"):
        OPTS = OSX_OPTS
    else:
        OPTS = LINUX_OPTS
else:
    OPTS = WINDOWS_OPTS


def build(setup_kwargs):
    setup_kwargs.update(
        {
            "ext_modules": [
                Extension(
                    "gamred_native",
                    sources=glob("gamred_native/*.c"),
                    include_dirs=["gamred_native", numpy.get_include()],
                    define_macros=[
                        ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
                    ],
                    **OPTS,
                ),
            ],
        }
    )



import logging

import numpy as np
from sklearn.base import BaseEstimator

import divik._matlab_legacy as ml
from divik.core import configurable

from ._stat_selector_mixin import StatSelectorMixin


@configurable
class GMMSelector(BaseEstimator, StatSelectorMixin):
    """Feature selector that removes low- or high- mean or variance features

    Gaussian Mixture Modeling is applied to the features' characteristics
    and components are obtained. Crossing points of the components are
    considered candidate thresholds. Out of these up to ``n_candidates``
    components are removed in such a way that at least ``min_features`` or
    ``min_features_rate`` features are retained.

    This feature selection algorithm looks only at the features (X), not the
    desired outputs (y), and can thus be used for unsupervised learning.

    Parameters
    ----------
    stat: {'mean', 'var', 'cv'}
        Kind of statistic to be computed out of the feature.

    neutral: float, optional, default: None
        This element will be omitted from the computation of the statistic.

    use_log: bool, optional, default: False
        Whether to use the logarithm of feature characteristic instead of the
        characteristic itself. This may improve feature filtering performance,
        depending on the distribution of features, however all the
        characteristics (mean, variance) have to be positive for that -
        filtering will fail otherwise. This is useful for specific cases in
        biology where the distribution of data may actually require this option
        for any efficient filtering.

    n_candidates: int, optional, default: None
        How many candidate thresholds to use at most. ``0`` preserves all the
        features (all candidate thresholds are discarded), ``None`` allows to
        remove all but one component (all candidate thresholds are retained).
        Negative value means to discard up to all but ``-n_candidates``
        candidates, e.g. ``-1`` will retain at least two components (one
        candidate threshold is removed).

    min_features: int, optional, default: 1
        How many features must be preserved. Candidate thresholds are tested
        against this value, and if they retain less features, less conservative
        thresholds is selected.

    min_features_rate: float, optional, default: 0.0
        Similar to ``min_features`` but relative to the input data features
        number.

    preserve_high: bool, optional, default: True
        Whether to preserve the high-characteristic features or
        low-characteristic ones.

    max_components: int, optional, default: 10
        The maximum number of components used in the GMM decomposition.

    Attributes
    ----------
    vals_: array, shape (n_features,)
        Computed characteristic of each feature.

    threshold_: float
        Threshold value to filter the features by the characteristic.

    raw_threshold_: float
        Threshold value mapped back to characteristic space (no logarithm, etc.)

    selected_: array, shape (n_features,)
        Vector of binary selections of the informative features.

    Examples
    --------
    >>> import numpy as np
    >>> import divik.feature_selection as fs
    >>> np.random.seed(42)
    >>> labels = np.concatenate([30 * [0] + 20 * [1] + 30 * [2] + 40 * [3]])
    >>> data = labels * 5 + np.random.randn(*labels.shape)
    >>> fs.GMMSelector('mean').fit_transform(data)
    array([[14.78032811 15.35711257 ... 15.75193303]])
    >>> fs.GMMSelector('mean', preserve_high=False).fit_transform(data)
    array([[ 0.49671415 -0.1382643  ... -0.29169375]])
    >>> fs.GMMSelector('mean', n_discard=-1).fit_transform(data)
    array([[10.32408397  9.61491772 ... 15.75193303]])
    """

    def __init__(
        self,
        stat: str,
        neutral: float = None,
        use_log: bool = False,
        n_candidates: int = None,
        min_features: int = 1,
        min_features_rate: float = 0.0,
        preserve_high: bool = True,
        max_components: int = 10,
    ):
        if stat not in {"cv", "mean", "var"} and not callable(stat):
            msg = 'stat must be one of {"cv", "mean", "var"} or callable'
            logging.error(msg)
            raise ValueError(msg)
        self.stat = stat
        self.neutral = neutral
        self.use_log = use_log
        self.n_candidates = n_candidates
        self.min_features = min_features
        self.min_features_rate = min_features_rate
        self.preserve_high = preserve_high
        self.max_components = max_components

    def fit(self, X, y=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.vals_ = self._to_characteristics(X)
        thrs = ml.find_thresholds(self.vals_, max_components=self.max_components)
        n_candidates = len(thrs) if self.n_candidates is None else self.n_candidates
        desired_thrs = thrs[:n_candidates]
        min_features = max(self.min_features, self.min_features_rate * X.shape[1])
        for thr in reversed(desired_thrs):
            selected = self.vals_ >= thr
            if selected.sum() >= min_features:
                break
        else:
            selected = np.ones((X.shape[1],), dtype=bool)
            thr = -np.inf
        self.threshold_ = thr
        self.raw_threshold_ = self._to_raw(thr)
        self.selected_ = selected
        return self



import logging
from functools import partial
from typing import Union

import numpy as np
from sklearn.base import clone

from divik.core import Data, maybe_pool
from divik.sampler import StratifiedSampler, UniformSampler
from divik.score._gap import _sampled_dispersion as _dispersion

KMeans = "divik.cluster.KMeans"
_BIG_PRIME = 40013


def _pool_initialize(initializers, *args):
    for arg, sampler in zip(args, initializers):
        sampler.initializer(*arg)


def sampled_gap(
    data: Data,
    kmeans: KMeans,
    sample_size: Union[int, float] = 1000,
    n_jobs: int = None,
    seed: int = 0,
    n_trials: int = 100,
    return_deviation: bool = False,
    reference_sampler=None,
) -> float:
    # TODO: Docs
    logging.debug("Creating samplers.")
    data_ = StratifiedSampler(n_rows=sample_size, n_samples=n_trials).fit(
        data, kmeans.labels_
    )
    if reference_sampler is None:
        reference_sampler = UniformSampler(n_rows=sample_size, n_samples=n_trials)
    reference_ = reference_sampler.fit(data)
    kmeans_ = clone(kmeans)
    seeds = list(seed + np.arange(n_trials) * _BIG_PRIME)
    logging.debug(f"Generated seeds: {seeds}.")
    logging.debug(f"Entering parallel context with n_jobs={n_jobs}.")
    with data_.parallel() as d, reference_.parallel() as r:
        initializer = partial(_pool_initialize, [d, r])
        with maybe_pool(
            n_jobs, initializer=initializer, initargs=(d.initargs, r.initargs)
        ) as pool:
            logging.debug("Computing reference dispersion.")
            compute_disp = partial(_dispersion, sampler=r, kmeans=kmeans_)
            ref_disp = pool.map(compute_disp, seeds)
            logging.debug("Computing data dispersion.")
            compute_disp = partial(_dispersion, sampler=d, kmeans=kmeans, fit=False)
            data_disp = pool.map(compute_disp, seeds)
    logging.debug("Left parallel context.")
    ref_disp = np.log(ref_disp)
    data_disp = np.log(data_disp)
    gap = np.mean(ref_disp) - np.mean(data_disp)
    result = (gap,)
    if return_deviation:
        std = np.sqrt(np.var(ref_disp) + np.var(data_disp)) / n_trials
        result += (std,)
    return result



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from itertools import count

from sklearn.base import BaseEstimator, clone


class BaseSampler(BaseEstimator, metaclass=ABCMeta):
    """Base class for all the samplers

    Sampler is Pool-safe, i.e. can simply store a dataset.
    It will not be serialized by pickle when going to another process,
    if handled properly.

    Before you spawn a pool, a data must be moved to a module-level
    variable. To simplify that process a contract has been prepared.
    You open a context and operate within a context:

    >>> with sampler.parallel() as sampler_,
    ...         Pool(initializer=sampler_.initializer,
    ...              initargs=sampler_.initargs) as pool:
    ...     pool.map(sampler_.get_sample, range(10))

    Keep in mind, that __iter__ and fit are not accessible in parallel
    context. __iter__ would yield the same values independently in
    all the workers. Now it needs to be done consciously and in
    well-though manner. fit could lead to a non-predictable behaviour.
    If you need the original sampler, you can get a clone (not fit to
    the data).
    """

    def __iter__(self):
        """Iter through `n_samples` samples or infinitely if unspecified"""
        if hasattr(self, "n_samples") and self.n_samples is not None:
            samples = range(self.n_samples)
        else:
            samples = count()
        for i in samples:
            yield self.get_sample(i)

    @abstractmethod
    def get_sample(self, seed):
        """Return specific sample

        Following assumptions should be met:
        a) sampler.get_sample(x) == sampler.get_sample(x)
        b) x != y should yield sampler.get_sample(x) != sampler.get_sample(y)

        Parameters
        ----------
        seed : int
            The seed to use to draw the sample

        Returns
        -------
        sample : array_like, (*self.shape_)
            Returns the drawn sample
        """
        raise NotImplementedError("get_sample is not implemented")

    def fit(self, X, y=None):
        """Fit sampler to data

        It's a base for both supervised and unsupervised samplers.
        """
        return self

    @contextmanager
    def parallel(self):
        """Create parallel context for the sampler to operate"""
        yield ParallelSampler(self)


class ParallelSampler:
    """Helper class for sharing the sampler functionality"""

    def __init__(self, sampler: BaseSampler):
        self.sampler = sampler

    def get_sample(self, seed):
        """Return specific sample"""
        return self.sampler.get_sample(seed)

    def initializer(self, *args):
        pass

    @property
    def initargs(self):
        return ()

    def clone(self):
        """Clones the original sampler"""
        return clone(self.sampler)



"""Reusable utilities used for building divik library"""

from ._cache import cached_fit
from ._gin_compat import (
    configurable,
    dump_gin_args,
    parse_args,
)
from ._parallel import (
    get_n_jobs,
    maybe_pool,
    share,
)
from ._seed import seed, seeded
from ._subsets import Subsets
from ._types import (
    Centroids,
    Data,
    DivikResult,
    IntLabels,
    SegmentationMethod,
)
from ._utils import (
    build,
    context_if,
    normalize_rows,
    visualize,
)

__all__ = [
    "cached_fit",
    "Centroids",
    "Data",
    "DivikResult",
    "IntLabels",
    "SegmentationMethod",
    "build",
    "context_if",
    "normalize_rows",
    "visualize",
    "get_n_jobs",
    "maybe_pool",
    "share",
    "seed",
    "seeded",
    "Subsets",
    "configurable",
    "dump_gin_args",
    "parse_args",
]



"""Clustering methods"""
from ._divik import DiviK
from ._kmeans import (
    DunnSearch,
    GAPSearch,
    KMeans,
)
from ._two_step import TwoStep

__all__ = [
    "DiviK",
    "DunnSearch",
    "GAPSearch",
    "KMeans",
    "TwoStep",
]



import logging

import gin

from divik.core import dump_gin_args, parse_args
from divik.core._utils import prepare_destination, setup_logger
from divik.core.io import (
    save,
    try_load_data,
    try_load_xy,
)


@gin.configurable
def load_data(path=gin.REQUIRED):
    return try_load_data(path)


@gin.configurable
def load_xy(path=None):
    return try_load_xy(path)


@gin.configurable
def experiment(
    model=gin.REQUIRED,
    steps_that_require_xy=None,
    destination: str = "result",
    omit_datetime: bool = False,
    verbose: bool = False,
    exist_ok: bool = False,
):
    destination = prepare_destination(destination, omit_datetime, exist_ok=exist_ok)
    dump_gin_args(destination)
    setup_logger(destination, verbose)
    logging.info(str(model))
    data = load_data()
    xy = load_xy()
    # repeated dump just because the dataset locations are not tracked
    dump_gin_args(destination)
    if steps_that_require_xy is None:
        steps_that_require_xy = []
    kwargs = {f"{step}__xy": xy for step in steps_that_require_xy}
    model.fit(data, **kwargs)
    save(model, destination, xy=xy)


def main():
    parse_args()
    experiment()


if __name__ == "__main__":
    main()



import numpy as np


def quantile(values, quantiles):
    """Compute MATLAB-alike quantiles

    Arguments:
        values - (np.ndarray) Input array or object that can be converted to an
        array.
        quantiles - (float, np.ndarray) float in range of [0,1] (or sequence
        of floats). Quantile to compute, which must be between 0 and 1
        inclusive.


    location of first element as a quantile -> 0.5 / n
    location of last element as a quantile -> (n - 0.5) / n

         ^
         |     (n-0.5) / n    1.0
     1.0 ----------------|---|
         |              /
         |             /
         |            /
         |           /                  <- this is how quantiles look in MATLAB
         |          /
         |         /
         |        /
         |       /
     0.0 ----|---|------------->
         | 0.0   0.5 / n


    y = ax + b


    0 = a * 0.5 / n + b         <- 0th element is treated as 0.5 / n quantile
    1 = a * (n - 0.5) / n + b   <- last element is treated as (n - 0.5) / n quantile

    a = n / (n - 1)
    b = - 0.5 / (n - 1)

    """
    values = np.array(values)
    n = float(values.size)
    a = n / (n - 1)
    b = -0.5 / (n - 1)
    quantiles = np.array(quantiles)
    matlab_alike_quantiles = np.clip(a * quantiles + b, a_min=0.0, a_max=1.0)
    return np.percentile(values, q=100.0 * matlab_alike_quantiles)


def n_quantiles(values, N, unbiased=True, backend=quantile):
    return backend(values, np.arange(1, N + 1, dtype=float) / (N + int(unbiased)))


def iqr(values, rng=(25, 75)):
    q1, q3 = quantile(values, 0.01 * np.array(rng))
    return q3 - q1



from typing import (
    Callable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import numpy as np

Table = np.ndarray  # 2D matrix
Data = Table
Centroids = Table
IntLabels = np.ndarray
SegmentationMethod = Callable[[Data], Tuple[IntLabels, Centroids]]
Clustering = Union["divik.cluster.GAPSearch", "divik.cluster.DunnSearch"]


class DivikResult(NamedTuple):
    """Result of DiviK clustering"""

    clustering: Clustering
    """Fitted automated clustering estimator"""
    feature_selector: "divik.feature_selection.StatSelectorMixin"
    """Fitted feature selector"""
    merged: IntLabels
    """Recursively merged clustering labels"""
    subregions: List[Optional["DivikResult"]]
    """DivikResults for all obtained subregions"""



import logging
import sys
import uuid
from functools import partial

import numpy as np
import tqdm
from sklearn.base import (
    BaseEstimator,
    ClusterMixin,
    TransformerMixin,
    clone,
)
from sklearn.utils.validation import check_is_fitted

from divik.core import (
    configurable,
    maybe_pool,
    share,
)
from divik.score import dunn, sampled_dunn

from ._core import KMeans

_DATA = {}


def _pool_initialize(ref, data):
    _DATA[ref] = data


@configurable
class DunnSearch(BaseEstimator, ClusterMixin, TransformerMixin):
    """Select best number of clusters for k-means

    Parameters
    ----------
    kmeans : KMeans
        KMeans object to tune

    max_clusters: int
        The maximal number of clusters to form and score.

    min_clusters: int, default: 1
        The minimal number of clusters to form and score.

    method: {'full', 'sampled', 'auto'}, default: 'full'
        Whether to run full computations or approximate.
        - full - always computes full Dunn's index, without sampling
        - sampled - samples the clusters to reduce computational overhead
        - auto - switches the above methods to provide best performance-quality
        trade-off.

    inter : {'centroid', 'closest'}, default: 'centroid'
        How the distance between clusters is computed. For more details see
        `dunn`.

    intra : {'avg', 'furthest'}, default: 'avg'
        How the cluster internal distance is computed. For more details see
        `dunn`.

    sample_size : int, default: 1000
        Size of the sample used to compute Dunn index in `auto` or `sampled`
        scenario.

    n_trials : int, default: 10
        Number of trials to use when computing Dunn index in `auto` or
        `sampled` scenario.

    seed : int, default: 42
        Random seed for the reproducibility of subset draws in Dunn `auto`
        or `sampled` scenario.

    n_jobs: int, default: 1
        The number of jobs to use for the computation. This works by computing
        each of the clustering & scoring runs in parallel.

    drop_unfit: bool, default: False
        If True, drops the estimators that did not fit the data.

    verbose: bool, default: False
        If True, shows progress with tqdm.

    Attributes
    ----------
    cluster_centers_: array, [n_clusters, n_features]
        Coordinates of cluster centers.

    labels_:
        Labels of each point.

    estimators_: List[KMeans]
        KMeans instances for n_clusters in range [min_clusters, max_clusters].

    scores_: array, [max_clusters - min_clusters + 1,]
        Array with scores for each estimator.

    n_clusters_: int
        Estimated optimal number of clusters.

    best_score_: float
        Score of the optimal estimator.

    best_: KMeans
        The optimal estimator.

    """

    def __init__(
        self,
        kmeans: KMeans,
        max_clusters: int,
        min_clusters: int = 2,
        method="full",
        inter="centroid",
        intra="avg",
        sample_size=1000,
        n_trials=10,
        seed=42,
        n_jobs: int = 1,
        drop_unfit: bool = False,
        verbose: bool = False,
    ):
        super().__init__()
        assert min_clusters <= max_clusters
        self.kmeans = kmeans
        self.min_clusters = min_clusters
        self.max_clusters = max_clusters
        self.method = method
        self.inter = inter
        self.intra = intra
        self.sample_size = sample_size
        self.n_trials = n_trials
        self.seed = seed
        self.n_jobs = n_jobs
        self.drop_unfit = drop_unfit
        self.verbose = verbose

    def _n_ops(self, data):
        if self.inter == "closest" or self.intra == "furthest":
            n_ops_full = data.shape[0] ** 2
            n_ops_sampled = self.n_trials * self.sample_size ** 2
        else:
            n_ops_full = data.shape[0]
            n_ops_sampled = self.n_trials * self.sample_size
        return n_ops_full, n_ops_sampled

    def _sampled_dunn(self, kmeans, data, inter, intra):
        return sampled_dunn(
            kmeans,
            data,
            inter=inter,
            intra=intra,
            sample_size=self.sample_size,
            n_jobs=1,
            seed=self.seed,
            n_trials=self.n_trials,
        )

    def _dunn(self, kmeans, data):
        n_ops_full, n_ops_sampled = self._n_ops(data)
        if self.method == "full":
            dunn_ = dunn
        elif self.method == "sampled":
            dunn_ = self._sampled_dunn
        elif self.method == "auto" and n_ops_full <= n_ops_sampled:
            dunn_ = dunn
            logging.debug('DunnSearch.method="auto" resolved to "full"')
        elif self.method == "auto" and n_ops_full > n_ops_sampled:
            dunn_ = self._sampled_dunn
            logging.debug('DunnSearch.method="auto" resolved to "sampled"')
        else:
            logging.error(f"Failed to select Dunn method {self.method}.")
            logging.debug(f"n_ops_full={n_ops_full}, " f"n_ops_sampled={n_ops_sampled}")
            raise ValueError(f"Unknown Dunn method {self.method}")
        return dunn_(kmeans, data, inter=self.inter, intra=self.intra)

    def _fit_kmeans(self, n_clusters, data_ref):
        data = _DATA[data_ref].value
        kmeans = clone(self.kmeans)
        kmeans.n_clusters = n_clusters
        kmeans.fit(data)
        d = self._dunn(kmeans, data)
        return kmeans, d

    def fit(self, X, y=None):
        """Compute k-means clustering and estimate optimal number of clusters.

        Parameters
        ----------

        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.

        y : Ignored
            not used, present here for API consistency by convention.

        """
        n_clusters = range(self.min_clusters, self.max_clusters + 1)
        if self.verbose:
            n_clusters = tqdm.tqdm(n_clusters, leave=False, file=sys.stdout)
        ref = str(uuid.uuid4())
        with share(X) as x:
            _DATA[ref] = x
            with maybe_pool(
                self.n_jobs, initializer=_pool_initialize, initargs=(ref, x)
            ) as pool:
                fit_kmeans = partial(self._fit_kmeans, data_ref=ref)
                kmeans_and_scores = pool.map(fit_kmeans, n_clusters)
            del _DATA[ref]
        logging.debug("Fitted DunnSearch")

        self.estimators_, self.scores_ = zip(*kmeans_and_scores)
        self.scores_ = np.array(self.scores_)
        best = np.argmax(self.scores_)
        self.fitted_ = True
        self.n_clusters_ = best + self.min_clusters
        self.best_score_ = self.scores_[best]
        self.best_ = self.estimators_[best]
        self.labels_ = self.best_.labels_
        self.cluster_centers_ = self.best_.cluster_centers_
        if self.drop_unfit:
            self.estimators_ = None

        return self

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------

        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)
        return self.best_.predict(X)

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------

        X_new : array, shape [n_samples, k]
            X transformed in the new space.

        """
        check_is_fitted(self)
        return self.best_.transform(X)



"""Reusable utilities for data and model I/O"""
from ._data_io import (
    load_data,
    save_csv,
    try_load_data,
    try_load_xy,
)
from ._model_io import save, saver

DIVIK_RESULT_FNAME = "result.pkl"

__all__ = [
    "load_data",
    "save_csv",
    "try_load_data",
    "try_load_xy",
    "save",
    "saver",
]



from functools import partial

import numpy as np
from skimage.exposure import cumulative_distribution
from sklearn.base import BaseEstimator, TransformerMixin

from divik.core import configurable, maybe_pool


@configurable
class HistogramEqualization(BaseEstimator, TransformerMixin):
    """Equalize histogram of the features to increase contrast
    
    Based on https://github.com/scikit-image/scikit-image/blob/master/skimage/exposure/exposure.py#L187-L223

    Parameters
    ----------

    n_bins : int, default 256
        Number of bins for histogram equalization.

    n_jobs : int, default -1
        Number of CPU cores to use during equalization

    Attributes
    ----------

    cdf_ : array
        Values of cumulative distribution function for all the features
    
    bins_ : array
        Bin centers for all the features
    """

    def __init__(self, n_bins: int = 256, n_jobs: int = -1):
        self.n_bins = n_bins
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        cdf = partial(cumulative_distribution, nbins=self.n_bins)
        with maybe_pool(processes=self.n_jobs) as pool:
            cdf_and_bins = pool.map(cdf, X.T)
        self.cdf_, self.bins_ = zip(*cdf_and_bins)

        return self

    def transform(self, X, y=None):
        features = zip(X.T, self.bins_, self.cdf_)
        with maybe_pool(processes=self.n_jobs) as pool:
            return np.transpose(pool.starmap(np.interp, features))



import logging
import sys
from functools import partial

import numpy as np
import tqdm
from sklearn.base import (
    BaseEstimator,
    ClusterMixin,
    TransformerMixin,
    clone,
)
from sklearn.utils.validation import check_is_fitted

from divik.cluster._kmeans._core import KMeans
from divik.core import configurable
from divik.sampler import BaseSampler
from divik.score import gap, sampled_gap

_BIG_PRIME = 32801


@configurable
class GAPSearch(BaseEstimator, ClusterMixin, TransformerMixin):
    """Select best number of cluters for k-means

    Parameters
    ----------
    kmeans : KMeans
        KMeans object to tune

    max_clusters: int
        The maximal number of clusters to form and score.

    min_clusters: int, default: 1
        The minimal number of clusters to form and score.

    reference_sampler: BaseSampler, default: None
        Sampler for reference dataset sampling in GAP statistic computations.

    n_jobs: int, default: 1
        The number of jobs to use for the computation. This works by computing
        each of the clustering & scoring runs in parallel.

    seed: int, default: 0
        Random seed for generating uniform data sets.

    n_trials: int, default: 10
        Number of data sets drawn as a reference.

    sample_size : int, default: 1000
        Size of the sample used for GAP statistic computation. Used only if
        introduces speedup.

    drop_unfit: bool, default: False
        If True, drops the estimators that did not fit the data.

    verbose: bool, default: False
        If True, shows progress with tqdm.

    Attributes
    ----------
    cluster_centers_: array, [n_clusters, n_features]
        Coordinates of cluster centers.

    labels_:
        Labels of each point.

    estimators_: List[KMeans]
        KMeans instances for n_clusters in range [min_clusters, max_clusters].

    scores_: array, [max_clusters - min_clusters + 1, ?]
        Array with scores for each estimator in each row.

    n_clusters_: int
        Estimated optimal number of clusters.

    best_score_: float
        Score of the optimal estimator.

    best_: KMeans
        The optimal estimator.

    """

    def __init__(
        self,
        kmeans: KMeans,
        max_clusters: int,
        min_clusters: int = 1,
        reference_sampler: BaseSampler = None,
        n_jobs: int = 1,
        seed: int = 0,
        n_trials: int = 10,
        sample_size: int = 1000,
        drop_unfit: bool = False,
        verbose: bool = False,
    ):
        super().__init__()
        assert min_clusters <= max_clusters
        self.kmeans = kmeans
        self.min_clusters = min_clusters
        self.max_clusters = max_clusters
        self.reference_sampler = reference_sampler
        self.n_jobs = n_jobs
        self.seed = seed
        self.n_trials = n_trials
        self.sample_size = sample_size
        self.drop_unfit = drop_unfit
        self.verbose = verbose

    def _should_sample(self, data):
        sampled_complexity = 2 * self.n_trials * self.sample_size ** 2
        normal_complexity = self.n_trials * data.shape[0] ** 2
        return sampled_complexity < normal_complexity

    def _gap(self, data, kmeans):
        if self._should_sample(data):
            logging.debug("Selecting sampled GAP.")
            score = partial(sampled_gap, sample_size=self.sample_size)
        else:
            logging.debug("Selecting full GAP.")
            score = gap
        return score(
            data,
            kmeans,
            n_jobs=self.n_jobs,
            seed=self.seed + _BIG_PRIME * kmeans.n_clusters,
            n_trials=self.n_trials,
            return_deviation=True,
            reference_sampler=self.reference_sampler,
        )

    def _fit_kmeans(self, n_clusters, data):
        kmeans = clone(self.kmeans)
        kmeans.n_clusters = n_clusters
        logging.debug(f"Fitting kmeans for {n_clusters} clusters.")
        kmeans.fit(data)
        logging.debug(f"Fitted kmeans for {n_clusters} clusters.")
        idx, std = self._gap(data, kmeans)
        logging.debug(f"GAP index: {idx}; std: {std}.")
        return kmeans, idx, std

    def fit(self, X, y=None):
        """Compute k-means clustering and estimate optimal number of clusters.

        Parameters
        ----------

        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.

        y : Ignored
            not used, present here for API consistency by convention.

        """
        fit_kmeans = partial(self._fit_kmeans, data=X)
        n_clusters = range(self.min_clusters, self.max_clusters + 1)
        if self.verbose:
            n_clusters = tqdm.tqdm(n_clusters, leave=False, file=sys.stdout)

        self.fitted_ = False
        self.estimators_ = []
        self.scores_ = []
        prev_gap = -np.inf
        for n_clust in n_clusters:
            kmeans, gap_, std = fit_kmeans(n_clust)
            if prev_gap > gap_ + std:
                self.fitted_ = True
                break
            prev_gap = gap_
            self.estimators_.append(kmeans)
            self.scores_.append((gap_, std))
        if self.verbose:
            n_clusters.close()
        self.scores_ = np.array(self.scores_)

        if self.fitted_:
            logging.debug("Fitted GAPSearch")
            self.best_ = self.estimators_[-1]
            self.best_score_ = self.scores_[-1]
            self.n_clusters_ = self.best_.n_clusters
            self.labels_ = self.best_.labels_
            self.cluster_centers_ = self.best_.cluster_centers_
            if self.drop_unfit:
                self.estimators_ = None
        else:
            logging.debug("Failed to fit GAPSearch")
            self.best_ = None
            self.best_score_ = None
            self.n_clusters_ = None
            self.labels_ = None
            self.cluster_centers_ = None

        return self

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------

        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)
        return self.best_.predict(X)

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------

        X_new : array, shape [n_samples, k]
            X transformed in the new space.

        """
        check_is_fitted(self)
        return self.best_.transform(X)



"""Access to MATLAB legacy functionalities."""
import faulthandler
from contextlib import contextmanager

import numpy as np

import gamred_native as gn


@contextmanager
def sigsegv_handler():
    was_enabled = faulthandler.is_enabled()
    faulthandler.enable()
    yield
    if not was_enabled:
        faulthandler.disable()


def find_thresholds(values: np.ndarray, max_components: int = 10) -> np.ndarray:
    """Find candidate thresholds for decomposition of values by GMM.

    Parameters
    ----------
    values:
        vector of values to decompose

    max_components:
        maximal number of components to decompose into

    Returns
    -------
    array of candidate thresholds from crossings between GMM components
    """
    if values.size <= max_components:
        max_components = values.size
    if values.size == 0:
        return np.array([])
    if max_components <= 0:
        raise ValueError("max_components must be positive")
    if values.ndim != 1:
        raise ValueError("values must be 1D array")
    values = np.ascontiguousarray(values)
    offset = np.min(values)
    if np.min(values) == np.max(values):
        return np.array([])
    with sigsegv_handler():
        return gn.find_thresholds(values - offset, max_components) + offset



import logging
from abc import ABCMeta

import numpy as np
from sklearn.base import BaseEstimator

try:
    from sklearn.feature_selection._base import SelectorMixin
except ModuleNotFoundError:
    from sklearn.feature_selection.base import SelectorMixin

from divik.core import configurable


class StatSelectorMixin(SelectorMixin, metaclass=ABCMeta):
    """
    Transformer mixin that performs feature selection given a support mask

    This mixin provides a feature selector implementation with ``transform`` and
    ``inverse_transform`` functionality given that ``selected_`` is specified
    during ``fit``.

    Additionally, provides a ``_to_characteristics`` and ``_to_raw`` implementations
    given ``stat``, optionally ``use_log`` and ``preserve_high``.
    """

    def _to_characteristics(self, X):
        """Extract & normalize characteristics from data"""
        try:
            neutral = self.neutral
        except AttributeError:
            neutral = None
        if neutral is not None:
            X = X.copy()
            X[X == neutral] = np.nan
        if self.stat == "mean":
            vals = np.nanmean(X, axis=0)
        elif self.stat == "var":
            vals = np.nanvar(X, axis=0)
        elif self.stat == "cv":
            vals = np.nanstd(X, axis=0) / np.nanmean(X, axis=0)
        elif callable(self.stat):
            vals = self.stat(X)
            if vals.size != X.shape[1]:
                raise RuntimeError(
                    "Computed statistic shape mismatch {0}".format(vals.shape)
                )
        else:
            msg = 'stat must be one of {"cv", "mean", "var"} or callable'
            logging.error(msg)
            raise ValueError(msg)

        if hasattr(self, "use_log") and self.use_log:
            if np.any(vals < 0):
                logging.error(
                    "Feature characteristic cannot be negative with log filtering"
                )
                raise ValueError(
                    "Feature characteristic cannot be negative with log filtering"
                )
            vals = np.log(vals)

        if hasattr(self, "preserve_high") and not self.preserve_high:
            vals = -vals

        return vals

    def _to_raw(self, threshold):
        """Convert threshold to the feature characteristic space"""
        if hasattr(self, "preserve_high") and not self.preserve_high:
            threshold = -threshold
        if hasattr(self, "use_log") and self.use_log:
            threshold = np.exp(threshold)
        return threshold

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.selected_


@configurable
class NoSelector(BaseEstimator, StatSelectorMixin):
    """Dummy selector to use when no selection is supposed to be made."""

    def __init__(self):
        pass

    def fit(self, X, y=None):
        """Pass data forward

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors to pass.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.selected_ = np.ones((X.shape[1],), dtype=bool)
        return self



from typing import NamedTuple

import numpy as np
from kneed import KneeLocator

FeaturesSelection = NamedTuple(
    "FeaturesSelection",
    [
        ("score", np.ndarray),
        ("index", np.ndarray),
        ("selection", np.ndarray),
        ("threshold", float),
    ],
)


def _gradient(values: np.ndarray, order: int = 1) -> np.ndarray:
    result = values
    for _ in range(order):
        result = np.gradient(result)
    return result


def _plateau_point(sorted_scores) -> int:
    gradient = np.abs(_gradient(sorted_scores))
    is_plateau = gradient <= np.percentile(gradient, 1)
    plateau_points = np.nonzero(is_plateau)[0]
    return int(np.median(plateau_points))


def _knee_point(decreasing_segment: np.ndarray) -> int:
    locator = KneeLocator(
        x=np.arange(decreasing_segment.size, dtype=int),
        y=decreasing_segment,
        S=1.0,
        curve="convex",
        direction="decreasing",
    )
    assert locator.knee is not None
    return locator.knee


def select_features(feature_scores) -> FeaturesSelection:
    feature_scores = feature_scores.ravel()
    index = np.argsort(-feature_scores)
    score = feature_scores[index]
    plateau_point = _plateau_point(score)
    concave_up_segment = score[:plateau_point]
    knee_location = _knee_point(concave_up_segment)
    threshold = score[knee_location]
    selection = feature_scores >= threshold
    return FeaturesSelection(score, index, selection, threshold)



"""gin-config compatibility"""
import os
import sys

try:
    import gin

    _HAS_GIN = True
except ImportError:
    _HAS_GIN = False


MISSING_GIN_ERROR = """
gin-config package missing. You should install divik with appropriate extras:
    pip install divik[gin]
"""


def dump_gin_args(destination):
    """Dump gin-config effective configuration

    If you have `gin` extras installed, you can call `dump_gin_args`
    save effective gin configuration to a file.
    """
    try:
        import gin
    except ImportError as ex:
        raise ImportError(MISSING_GIN_ERROR) from ex
    with open(os.path.join(destination, "config.gin"), "w") as outfile:
        outfile.write(gin.operative_config_str())


if _HAS_GIN:
    configurable = gin.configurable
else:

    def configurable(name_or_fn=None, *args, **kwargs):
        if name_or_fn is None:
            return lambda x: x
        return name_or_fn


def parse_args():
    """Parse gin config files and parameter overrides from command line"""
    import argparse

    import gin

    parser = argparse.ArgumentParser()

    parser.add_argument("--param", nargs="*", help="List of Gin parameter bindings")
    parser.add_argument("--config", nargs="*", help="List of paths to the config files")

    args = parser.parse_args()

    gin.parse_config_files_and_bindings(args.config, args.param)



from functools import wraps

import joblib

from ._gin_compat import configurable


@configurable
def cache_path(path: str = "cache"):
    # This is required to make the cache path configurable
    return path


def _is_computed(fieldname):
    return (
        fieldname.endswith("_")
        and not fieldname.endswith("__")
        and not fieldname.startswith("_")
    )


def cached_fit(cls):
    """Decorate a sklearn-compatible estimator to cache the fitting result

    It is a wrapper over joblib.Memory.cache, that supports runtime cache
    path definition.

    Set path definition through gin config with ``cache_path.path``
    identifier.
    """
    _fit = cls.fit

    @wraps(_fit)
    def fit(self, X, y=None):
        # This is pushed inside for the sake of parameters parsing
        memory = joblib.Memory(location=cache_path())
        cached = memory.cache(_fit)
        output = cached(self, X, y)
        self.set_params(**output.get_params())
        computed_fields = [f for f in dir(output) if _is_computed(f)]
        for field in computed_fields:
            setattr(self, field, getattr(output, field))
        return self

    cls.fit = fit
    return cls



import logging
from typing import Tuple, Union

import dask.array as da
import dask.dataframe as dd
import dask_distance as ddst
import numpy as np
import scipy.spatial.distance as dst
from sklearn.base import (
    BaseEstimator,
    ClusterMixin,
    TransformerMixin,
)
from sklearn.utils.validation import check_is_fitted

from divik.cluster._kmeans._initialization import (
    ExtremeInitialization,
    Initialization,
    KDTreeInitialization,
    KDTreePercentileInitialization,
    PercentileInitialization,
)
from divik.core import (
    Centroids,
    Data,
    IntLabels,
    SegmentationMethod,
    configurable,
    normalize_rows,
)


class Labeling(object):
    """Labels observations by closest centroids"""

    def __init__(self, distance_metric: str, allow_dask: bool = False):
        """
        @param distance_metric: distance metric for estimation of closest
        @param allow_dask: should be False if `multiprocessing.Pool` is spawned
        """
        self.distance_metric = distance_metric
        self.allow_dask = allow_dask

    def __call__(self, data: Data, centroids: Centroids) -> IntLabels:
        """Find closest centroids

        @param data: observations in rows
        @param centroids: centroids in rows
        @return: vector of labels of centroids closest to points
        """
        if data.shape[1] != centroids.shape[1]:
            msg = (
                "Dimensionality of data and centroids must be equal. "
                + f"Was {data.shape[1]} and {centroids.shape[1]}"
            )
            logging.error(msg)
            raise ValueError(msg)

        if self.allow_dask and (data.shape[0] > 10000 or data.shape[1] > 1000):
            X1 = da.from_array(data)
            X2 = da.from_array(centroids)
            distances = ddst.cdist(X1, X2, self.distance_metric)
            labels = da.argmin(distances, axis=1).compute()
        else:
            distances = dst.cdist(data, centroids, self.distance_metric)
            labels = np.argmin(distances, axis=1)
        return labels


def redefine_centroids(
    data: Data, labeling: IntLabels, label_set: IntLabels, allow_dask: bool = False
) -> Centroids:
    """Recompute centroids in data for given labeling

    @param data: observations
    @param labeling: partition of dataset into groups
    @param label_set: set of labels used for partitioning
    @param allow_dask: should be False if `multiprocessing.Pool` is spawned
    @return: centroids
    """
    if data.shape[0] != labeling.size:
        msg = (
            "Each observation must have label specified. Number "
            + f"of labels: {labeling.size}, "
            + f"number of observations: {data.shape[0]}."
        )
        logging.error(msg)
        raise ValueError(msg)
    if allow_dask and (data.shape[0] > 10000 or data.shape[1] > 1000):
        X = dd.from_array(data)
        y = dd.from_array(labeling)
        centroids = X.groupby(y).mean().compute().values
    else:
        centroids = np.nan * np.zeros((len(label_set), data.shape[1]))
        for label in label_set:
            centroids[label] = np.mean(data[labeling == label], axis=0)
    return centroids


def _validate_kmeans_input(data: Data, number_of_clusters: int):
    if not isinstance(data, np.ndarray) or len(data.shape) != 2:
        logging.error("data is expected to be 2D np.array")
        raise ValueError("data is expected to be 2D np.array")
    if number_of_clusters < 1:
        msg = "number_of_clusters({0}) < 1".format(number_of_clusters)
        logging.error(msg)
        raise ValueError(msg)


def _validate_normalizable(data):
    is_constant = data.min(axis=1) == data.max(axis=1)
    if is_constant.any():
        constant_rows = np.where(is_constant)[0]
        msg = "Constant rows {0} are not allowed for normalization."
        logging.error(msg.format(constant_rows))
        raise ValueError(msg.format(constant_rows))


class _KMeans(SegmentationMethod):
    """K-means clustering"""

    def __init__(
        self,
        labeling: Labeling,
        initialize: Initialization,
        number_of_iterations: int = 100,
        normalize_rows: bool = False,
        allow_dask: bool = False,
    ):
        """
        @param labeling: labeling method
        @param initialize: initialization method
        @param number_of_iterations: number of iterations
        @param normalize_rows: sets mean of row to 0 and norm to 1
        @param allow_dask: should be False if `multiprocessing.Pool` is spawned
        """
        self.labeling = labeling
        self.initialize = initialize
        self.number_of_iterations = number_of_iterations
        self.normalize_rows = normalize_rows
        self.allow_dask = allow_dask

    def _fix_labels(self, data, centroids, labels, n_clusters, retries=10):
        logging.debug("A label vanished - fixing")
        new_labels = labels.copy()
        known_labels = np.unique(labels)
        expected_labels = np.arange(n_clusters)
        missing_labels = np.setdiff1d(expected_labels, known_labels)
        logging.debug(
            "Missing labels ({0} were expected): {1}".format(n_clusters, missing_labels)
        )
        new_centroids = np.nan * np.zeros((n_clusters, centroids.shape[1]))
        for known in known_labels:
            new_centroids[known] = centroids[known]
        for missing in missing_labels:
            logging.debug("Fixing label: {0}".format(missing))
            new_center = np.nanmin(
                dst.cdist(data, new_centroids, metric=self.labeling.distance_metric),
                axis=1,
            ).argmax()
            logging.debug("Assigning to label: {0}".format(labels[new_center]))
            new_labels[new_center] = missing
            new_centroids[missing] = data[new_center]
        if np.unique(new_labels).size != n_clusters and retries > 0:
            logging.debug("fixed but lost another: {0}".format(np.unique(new_labels)))
            return self._fix_labels(
                data, new_centroids, new_labels, n_clusters, retries - 1
            )
        return new_centroids, new_labels

    def __call__(
        self, data: Data, number_of_clusters: int
    ) -> Tuple[IntLabels, Centroids]:
        _validate_kmeans_input(data, number_of_clusters)
        if number_of_clusters == 1:
            return (
                np.zeros((data.shape[0], 1), dtype=int),
                np.mean(data, axis=0, keepdims=True),
            )
        data = data.reshape(data.shape, order="C")
        if self.normalize_rows:
            _validate_normalizable(data)
            data = normalize_rows(data)
        label_set = np.arange(number_of_clusters)
        logging.debug("Initializing KMeans centroids.")
        centroids = self.initialize(data, number_of_clusters)
        logging.debug("First centroids found.")
        old_labels = np.nan * np.zeros((data.shape[0],))
        labels = self.labeling(data, centroids)
        logging.debug("Labels assigned.")
        for _ in range(self.number_of_iterations):
            if np.unique(labels).size != number_of_clusters:
                centroids, labels = self._fix_labels(
                    data, centroids, labels, number_of_clusters
                )
            if np.all(labels == old_labels):
                logging.debug("Stability achieved.")
                break
            old_labels = labels
            centroids = redefine_centroids(data, old_labels, label_set, self.allow_dask)
            labels = self.labeling(data, centroids)
        return labels, centroids


def _parse_initialization(
    name: str,
    distance: str,
    percentile: float = None,
    leaf_size: Union[int, float] = 0.01,
) -> Initialization:
    if name == "percentile":
        return PercentileInitialization(distance, percentile)
    if name == "extreme":
        return ExtremeInitialization(distance)
    if name == "kdtree":
        return KDTreeInitialization(distance, leaf_size)
    if name == "kdtree_percentile":
        return KDTreePercentileInitialization(distance, leaf_size, percentile)
    logging.error("Unknown initialization: {0}".format(name))
    raise ValueError("Unknown initialization: {0}".format(name))


@configurable
class KMeans(BaseEstimator, ClusterMixin, TransformerMixin):
    """K-Means clustering

    Parameters
    ----------

    n_clusters : int
        The number of clusters to form as well as the number of
        centroids to generate.

    distance : str, optional, default: 'euclidean'
        Distance measure. One of the distances supported by scipy package.

    init : {'percentile', 'extreme', 'kdtree', 'kdtree_percentile'}
        Method for initialization, defaults to 'percentile':

        'percentile' : selects initial cluster centers for k-mean
        clustering starting from specified percentile of distance to
        already selected clusters

        'extreme': selects initial cluster centers for k-mean
        clustering starting from the furthest points to already specified
        clusters

        'kdtree': selects initial cluster centers for k-mean clustering
        starting from centroids of KD-Tree boxes

        'kdtree_percentile': selects initial cluster centers for k-means
        clustering starting from centroids of KD-Tree boxes containing
        specified percentile. This should be more robust against outliers.

    percentile : float, default: 95.0
        Specifies the starting percentile for 'percentile' initialization.
        Must be within range [0.0, 100.0]. At 100.0 it is equivalent to
        'extreme' initialization.

    leaf_size : int or float, optional (default 0.01)
        Desired leaf size in kdtree initialization. When int, the box size
        will be between `leaf_size` and `2 * leaf_size`. When float, it will
        be between `leaf_size * n_samples` and `2 * leaf_size * n_samples`

    max_iter : int, default: 100
        Maximum number of iterations of the k-means algorithm for a
        single run.

    normalize_rows : bool, default: False
        If True, rows are translated to mean of 0.0 and scaled to norm of 1.0.

    allow_dask : bool, default: False
        If True, automatically selects dask as computations backend whenever
        reasonable. Default `False` since it cannot be used together with
        `multiprocessing.Pool` and everywhere `n_jobs` must be set to `1`.

    Attributes
    ----------

    cluster_centers_ : array, [n_clusters, n_features]
        Coordinates of cluster centers.

    labels_ :
        Labels of each point

    """

    # TODO: Add example of usage.
    def __init__(
        self,
        n_clusters: int,
        distance: str = "euclidean",
        init: str = "percentile",
        percentile: float = 95.0,
        leaf_size: Union[int, float] = 0.01,
        max_iter: int = 100,
        normalize_rows: bool = False,
        allow_dask: bool = False,
    ):
        super().__init__()
        self.n_clusters = n_clusters
        self.distance = distance
        self.init = init
        self.percentile = percentile
        self.leaf_size = leaf_size
        self.max_iter = max_iter
        self.normalize_rows = normalize_rows
        self.allow_dask = allow_dask

    def fit(self, X, y=None):
        """Compute k-means clustering.

        Parameters
        ----------

        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.

        y : Ignored
            not used, present here for API consistency by convention.
        """
        initialize = _parse_initialization(
            self.init, self.distance, self.percentile, self.leaf_size
        )
        kmeans = _KMeans(
            labeling=Labeling(self.distance, allow_dask=self.allow_dask),
            initialize=initialize,
            number_of_iterations=self.max_iter,
            normalize_rows=self.normalize_rows,
            allow_dask=self.allow_dask,
        )
        X = np.asanyarray(X)
        self.labels_, self.cluster_centers_ = kmeans(
            X, number_of_clusters=self.n_clusters
        )
        self.labels_ = self.labels_.ravel()
        return self

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------

        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)
        if self.normalize_rows:
            X = normalize_rows(X)
        labels = dst.cdist(X, self.cluster_centers_, self.distance).argmin(axis=1)
        return labels

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------

        X_new : array, shape [n_samples, k]
            X transformed in the new space.

        """
        check_is_fitted(self)
        if self.normalize_rows:
            X = normalize_rows(X)
        return dst.cdist(X, self.cluster_centers_, self.distance)



import logging
from functools import partial

import numpy as np
import pandas as pd
import scipy.spatial.distance as dist
from sklearn.base import clone

from divik.core import (
    Data,
    maybe_pool,
    normalize_rows,
    seeded,
)
from divik.sampler import BaseSampler, UniformSampler

KMeans = "divik.KMeans"
_BIG_PRIME = 54673


def _dispersion(data: Data, kmeans: KMeans) -> float:
    assert data.shape[0] == kmeans.labels_.size, "kmeans not fit on this data"
    if getattr(kmeans, "normalize_rows", False):
        data = normalize_rows(data)
    clusters = pd.DataFrame(data).groupby(kmeans.labels_)
    return float(
        np.mean(
            [
                np.mean(dist.pdist(cluster_members.values, kmeans.distance))
                for _, cluster_members in clusters
                if cluster_members.shape[0] != 1
            ]
        )
    )


def _sampled_dispersion(
    seed: int, sampler: BaseSampler, kmeans: KMeans, fit: bool = True
) -> float:
    logging.debug(f"Sampling with seed {seed}.")
    X = sampler.get_sample(seed)
    logging.debug(f"Sample shape {X.shape}")
    if getattr(kmeans, "normalize_rows", False):
        logging.debug("Normalizing rows.")
        X = normalize_rows(X)
    if fit:
        logging.debug("Fitting kmeans for sample.")
        y = kmeans.fit_predict(X)
    else:
        logging.debug("Predicting labels for sample.")
        y = kmeans.predict(X)
    logging.debug("Computing dispersion for clustered sample.")
    clusters = pd.DataFrame(X).groupby(y)
    return float(
        np.mean(
            [
                np.mean(dist.pdist(cluster_members.values, kmeans.distance))
                for _, cluster_members in clusters
                if cluster_members.shape[0] != 1
            ]
        )
    )


def gap(
    data: Data,
    kmeans: KMeans,
    n_jobs: int = None,
    seed: int = 0,
    n_trials: int = 100,
    return_deviation: bool = False,
    reference_sampler=None,
) -> float:
    if reference_sampler is None:
        reference_sampler = UniformSampler(n_rows=None, n_samples=n_trials)
    reference_ = reference_sampler.fit(data)
    kmeans_ = clone(kmeans)
    seeds = list(seed + np.arange(n_trials) * _BIG_PRIME)
    with reference_.parallel() as r:
        with maybe_pool(n_jobs, initializer=r.initializer, initargs=r.initargs) as pool:
            compute_disp = partial(_sampled_dispersion, sampler=r, kmeans=kmeans_)
            ref_disp = pool.map(compute_disp, seeds)
    ref_disp = np.log(ref_disp)
    data_disp = np.log(_dispersion(data, kmeans))
    gap = np.mean(ref_disp) - data_disp
    result = (gap,)
    if return_deviation:
        std = np.sqrt(1 + 1 / n_trials) * np.std(ref_disp)
        result += (std,)
    return result
